#include "Python.h"
#include <sys/time.h>

static PyObject *ItimerError;

static void
timeval_from_double(double d, struct timeval *tv)
{
    tv->tv_sec = floor(d);
    tv->tv_usec = fmod(d, 1.0) * 1000000.0;
}

static inline double
double_from_timeval(struct timeval *tv)
{
    return tv->tv_sec + (double)(tv->tv_usec / 1000000.0);
}

static PyObject *
itimer_retval(struct itimerval *iv)
{
    PyObject *r,*v;

    if(!(r = PyTuple_New(2)))
        return NULL;

    if(!(v = PyFloat_FromDouble(double_from_timeval(&iv->it_value))))
        return NULL;

    PyTuple_SET_ITEM(r, 0, v);

    if(!(v = PyFloat_FromDouble(double_from_timeval(&iv->it_interval))))
        return NULL;

    PyTuple_SET_ITEM(r, 1, v);

    return r;
}

static PyObject *
itimer_setitimer(PyObject *self, PyObject *args)
{
    double first;
    double interval = 0;
    int which;
    struct itimerval new, old;

    if(!PyArg_ParseTuple(args, "id|d:setitimer", &which, &first, &interval))
        return NULL;

    /* Let OS do checking on which */
    timeval_from_double(first, &new.it_value);
    timeval_from_double(interval, &new.it_interval);
    if (setitimer(which, &new, &old) != 0) {
        PyErr_SetFromErrno(ItimerError);
        return NULL;
    }

    return itimer_retval(&old);
}

PyDoc_STRVAR(setitimer_doc,
"setitimer(which, seconds[, interval])\n\
\n\
Sets given itimer to fire after value seconds and after that every \n\
interval seconds.\n\
Clear itimer by setting seconds to zero.\n\
\n\
Returns old values as a tuple: (delay, interval)."); // verificar se eh tupla

static PyObject *
itimer_getitimer(PyObject *self, PyObject *args)
{
    int which;
    struct itimerval old;

    if(!PyArg_ParseTuple(args, "i:getitimer", &which))
        return NULL;

    if(getitimer(which, &old) != 0) {
        PyErr_SetFromErrno(ItimerError);
        return NULL;
    }

    return itimer_retval(&old);
}

PyDoc_STRVAR(getitimer_doc,
"getitimer(which)\n\
\n\
Returns current value of given itimer");

/* List of functions defined in the module */
static PyMethodDef itimer_methods[] = {
	{"setitimer",   itimer_setitimer,   METH_VARARGS, setitimer_doc},
	{"getitimer",   itimer_getitimer,   METH_VARARGS, getitimer_doc},
	{NULL,          NULL}		/* sentinel */
};


PyDoc_STRVAR(module_doc,
"This module wraps setitimer(2) and getitimer(2) unix syscalls. Those\n\
functions can be used as to implement subsecond alarm signal delivery.\n\
\n\
Functions:\n\
\n\
setitimer() -- setitimer to fire\n\
getitimer() -- get current value of timer\n\
\n\
Constants:\n\
\n\
ITIMER_REAL -- decrements in real time, and delivers SIGALRM upon\n\
               expiration\n\
ITIMER_VIRTUAL -- decrements only when the process is executing,\n\
               and delivers SIGVTALRM upon expiration\n\
ITIMER_PROF -- decrements both when the process is executing and\n\
               when the system is executing on behalf of the process.\n\
               Coupled with ITIMER_VIRTUAL, this timer is usually\n\
               used to profile the time spent by the application\n\
               in user and kernel space. SIGPROF is delivered upon\n\
               expiration.");

PyMODINIT_FUNC
inititimer(void)
{
    PyObject *m, *d;
    PyObject *v;

    /* Create the module and add the functions */
    m = Py_InitModule3("itimer", itimer_methods, module_doc);
    if (m == NULL)
        return;

    /* Add some symbolic constants to the module */
    d = PyModule_GetDict(m);
    ItimerError = PyErr_NewException("itimer.error", NULL, NULL);

#ifdef ITIMER_REAL
    v=PyInt_FromLong(ITIMER_REAL);
    PyDict_SetItemString(d, "ITIMER_REAL", v);
    Py_DECREF(v);
#endif
#ifdef ITIMER_VIRTUAL
    v=PyInt_FromLong(ITIMER_VIRTUAL);
    PyDict_SetItemString(d, "ITIMER_VIRTUAL",v);
    Py_DECREF(v);
#endif
#ifdef ITIMER_PROF
    v=PyInt_FromLong(ITIMER_PROF);
    PyDict_SetItemString(d, "ITIMER_PROF",v);
    Py_DECREF(v);
#endif

    PyDict_SetItemString(d, "error", ItimerError);
}
